#include "parser.hpp"
#include "visitors/debug_visitor.hpp"
#include "visitors/find_t_visitor.hpp"

#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>
#include <cstdint>
#include <sha1/sha1.hpp>
#include <sha1/BufferSource.hpp>

namespace {
std::string load_file (const std::string& path) {
	std::ifstream ifs(path);
	ifs >> std::noskipws;

	return std::string(
		std::istreambuf_iterator<char>(ifs),
		std::istreambuf_iterator<char>()
	);
}

SHA1::MessageDigest to_hash160_digest (const std::array<std::uint32_t, 5>& arr) {
	SHA1::MessageDigest hash_found;
	std::copy_n(arr.begin(), 5, hash_found.hash);
	return hash_found;
}
} //unnamed namespace

int main(int argc, const char* argv[]) {
	if (argc != 2) {
		std::cerr << "Wrong number of parameters. Usage:\n"
			<< argv[0] << " <path_to_torrent>\n";
		return 2;
	}

	std::string full_torrent = load_file(argv[1]);
	std::cout << "Loaded file into string of size " << full_torrent.size() << '\n';
	std::vector<duck::TorrentValue> values = duck::parse_torrent(full_torrent);

	for (const auto& value : values) {
		duck::DebugVisitor visitor;
		boost::apply_visitor(visitor, value);
	}

	auto hashes = duck::collect_hashes(values);
	std::cout << "Got " << hashes.size() << " hashes\n";

	const auto file_count = duck::find_int("/info/files/[[size]]", values);
	if (file_count)
		std::cout << "Input has " << file_count << " file entries\n";
	else
		std::cout << "Input seems to contain a single file only\n";

	if (0 == file_count) {
		std::string file_name = std::string{duck::find_string("/info/name", values)};
		std::cout << "Found file name \"" << file_name << "\"\n";

		const auto piece_length = duck::find_int("/info/piece length", values);
		std::ifstream istream(file_name, std::ios::in|std::ios::binary);

		if (istream.is_open()) {
			std::vector<std::uint8_t> buff(piece_length);
			std::size_t hash_index = 0;
			std::size_t match_count = 0;

			while (istream.read(reinterpret_cast<char*>(buff.data()), piece_length).gcount() > 0) {
				SHA1::BufferSource buff_source(buff.data(), istream.gcount() * CHAR_BIT);
				SHA1::MessageDigest calculated_hash = SHA1::computeFromSource(buff_source);
				SHA1::MessageDigest stored_hash = to_hash160_digest(hashes[hash_index]);
				std::cout << stored_hash.toHexString() << "  " <<
					calculated_hash.toHexString() << "  " << file_name << '\n';
				++hash_index;
				if (stored_hash == calculated_hash)
					++match_count;
			}

			std::cout << "Hash verified " << match_count << '/' << hashes.size() << '\n';
		}
		else {
			std::cout << "Error: opening \"" << file_name << "\" failed\n";
		}
	}
	else {
		std::string search("/info/files/[[");
		for (std::int_fast32_t z = 0; z < static_cast<std::int_fast32_t>(file_count); ++z) {
			const auto path = duck::find_string(search + std::to_string(z) + "]]/path/[[0]]", values);
			std::cout << "path index " << z << '\t' << path << '\n';
		}
	}
}
