#include "debug_visitor.hpp"

#include <iostream>
#include <ciso646>
#include <boost/variant/apply_visitor.hpp>
#include <algorithm>
#include <cctype>

namespace duck {
namespace {
bool is_printable(std::string_view s) {
    return std::find_if(s.cbegin(), s.cend(),
		 [](unsigned char c){ return not std::isprint(c); }
	) == s.cend();
}

} //unnamed namespace

DebugVisitor::DebugVisitor(int depth, bool skip_first) :
	m_depth(depth), m_first_line(skip_first), m_skip_first(skip_first) {}

void DebugVisitor::reset() {
	m_first_line = m_skip_first;
}

void DebugVisitor::operator() (TorrentIntType value) {
	print_tabs();
	std::cout << value << '\n';
}

void DebugVisitor::operator() (const TorrentStringType& value) {
	constexpr std::size_t max_size = 41;

	print_tabs();

	std::string_view str{value};
	std::string_view ellip;
	if (is_printable(value)) {
		if (value.size() > max_size + 3) {
			str = str.substr(0, max_size);
			ellip = "...";
		}
	}
	else {
		str = "<binary_data>";
	}

	std::cout << "\"" << str << ellip << "\"[" << value.size() << "]\n";
}

void DebugVisitor::operator() (const std::vector<TorrentValue>& value) {
	print_tabs();
	std::cout << "[\n";

	DebugVisitor new_visitor(m_depth+1, false);
	for (const auto& item : value) {
		boost::apply_visitor(new_visitor, item);
	}

	print_tabs();
	std::cout << "]\n";
}

void DebugVisitor::operator() (const std::map<TorrentStringType, TorrentValue>& value) {
	print_tabs();
	std::cout << "{\n";

	DebugVisitor new_visitor(m_depth+1);
	for (const auto& item : value) {
		print_tabs(1);
		std::cout << item.first << " => ";

		new_visitor.reset();
		boost::apply_visitor(new_visitor, item.second);
	}
	print_tabs();
	std::cout << "}\n";
}

void DebugVisitor::print_tabs(int extra) {
	if (!m_first_line) {
		for (int z = 0; z < m_depth+extra; ++z) std::cout << '\t';
	}
	else {
		m_first_line = false;
	}
}
} //namespace duck
