#pragma once

#include "../parser.hpp"
#include <string_view>
#include <cstddef>
#include <boost/variant/apply_visitor.hpp>
#include <vector>
#include <ciso646>

namespace duck {
namespace detail {
template <typename T>
class FindTVisitor : boost::static_visitor<T> {
public:
	FindTVisitor (std::string_view search_path);

	void reset();
	bool has_error() const noexcept;
	T operator() (TorrentIntType value);
	T operator() (const TorrentStringType& value);
	T operator() (const std::vector<TorrentValue>&);
	T operator() (const std::map<TorrentStringType, TorrentValue>& value);

private:
	bool is_path_found() const;

	std::vector<std::string> m_search;
	std::size_t m_search_match_count{0};
	bool m_search_error{false};
};
} //namespace detail

typedef detail::FindTVisitor<TorrentStringType> FindStringVisitor;
typedef detail::FindTVisitor<TorrentIntType> FindIntVisitor;

template <typename T>
inline T find_item(std::string_view path, const std::vector<TorrentValue>& values) {
	detail::FindTVisitor<T> visitor(path);
	for (const auto& value : values) {
		visitor.reset();
		T found = boost::apply_visitor(visitor, value);
		if (not visitor.has_error())
			return found;
	}
	return {};
}

inline TorrentIntType find_int(std::string_view path, const std::vector<TorrentValue>& values) {
	return find_item<TorrentIntType>(path, values);
}

inline TorrentStringType find_string(std::string_view path, const std::vector<TorrentValue>& values) {
	return find_item<TorrentStringType>(path, values);
}
} //namespace duck
